(*pp ./tracepp *)

open Vocable

let start_time = ref None
let line_width = 79

let user_exit _ = 
  print_newline ();
  print_newline ();
  Statistics.display_answer_stats ();
  print_newline ();
  (match !start_time with
       None -> ()
     | Some t0 ->
         let t1 = Time.time () in
         let s = Utils.string_of_tdiff (Int32.sub t1 t0) in
           print_string "Duration of session: ";
           print_endline s;
  );
  print_endline "Goodbye!";
  exit 0

let get_line () = 
  try
    read_line ()
  with End_of_file ->
    user_exit ()

let print_para ?(first_line_indent=true) indent words = 
  let istr = String.make indent ' ' in
  let rec print words lwidth after_lbreak = 
    match words with 
        [] -> print_newline ()
      | w :: rest ->
        let n = String.length w in
          if n <= lwidth || after_lbreak then (
            let nprinted = 
              if not after_lbreak then (
                print_string " "; 
                print_string w;
                n + 1
              ) else (
                print_string w;
                n
              ) 
            in
              print rest (lwidth - nprinted) false
          ) else (
            print_newline ();
            print_string istr;
            print words (line_width - indent) true
          )
  in
    (if first_line_indent then
       print_string istr
     else ());
    print words (line_width - indent) true

let play_audio_file fname = 
  if not (Sys.file_exists fname) then
    Printf.printf "Audio file '%s' does not exist\n" fname
  else
    let cmd = 
      if Filename.check_suffix fname ".mp3" then
        Some (Config.get_mp3_player ())
      else if Filename.check_suffix fname ".wav" then
        Some (Config.get_wav_player ())
      else (
        Printf.printf "Unknown audio format: %s\n" fname;
        None
      )
    in
      match cmd with
          None -> ()
        | Some cmd ->
            let cmdline = cmd ^ " " ^ fname in
              ITRACE "Playing audio file, command: '%s'" cmdline;
              print_endline 
"If you have an other sound program running, it may block playing the
pronunciation. In such a case, just stop the other program.";
              let ecode = Sys.command cmdline in
                if ecode = 0 then
                  ITRACE "Command finished with exit code 0"
                else
                  Printf.printf "Command failed, exit code: %d\n" ecode
                    

(* ==================== Commands ================================ *)

let print_all_trace_levels () =
  let names = Trace.level_names in
    List.iter (fun name ->
                 let l = Trace.level_of_string name in
                   Printf.printf "%s: %s\n" name
                     (if Trace.is_enabled l then "enabled"
                      else "disabled"))
      names

let adjust_tracing _ _ _ s =
  if s = "" then
    print_all_trace_levels ()
  else
    try
      let level = Trace.level_of_string s in
        Trace.toggle_level level
    with Invalid_argument _ -> (
      Printf.printf "Invalid level: '%s'\n" s;
      print_all_trace_levels ()
    )
  
let play_pronunciation ses voc before_answer _ = 
  match 
    if before_answer then voc.Vocable.voc_termPronunciation 
    else voc.Vocable.voc_meaningPronunciation 
  with
      Some simple_name ->
        let fname = Session.get_audio_filename ses simple_name in
          play_audio_file fname
    | None -> print_endline "No pronunciation available."

let rec display_help _ _ _ _ =
  let rec f (name, _, text) = 
    Printf.printf "%s    %s\n" name text
  in
    print_endline "The following commands are available:";
    List.iter f commands

and commands = 
  [(":?", display_help, "Display this help message.");
   (":q", (fun _ _ _ _ -> user_exit ()), "Exit the program.");
   (":t", adjust_tracing, 
    "Toggle given trace level. Without argument: print list of trace levels");
   (":s", (fun _ voc _ _ -> Statistics.display_voc_stats voc),
    "Display statistics for the current vocable");
   (":g", (fun ses _ _ _ -> Statistics.display_global_stats ses),
    "Display statistics for all vocables");
   (":p", play_pronunciation, 
    "Play pronunciation audio file for the current vocable.")
  ]


let exec_cmd session voc before_answer cmd = 
  let rec find_cmd l =
    match l with
        [] -> false
      | (c, f, _) :: rest ->
          if StrExtras.prefix c cmd then
            let args = Utils.strip (StrExtras.cut_first_n cmd 
                                          (String.length c)) in
              f session voc before_answer args;
              true
          else
            find_cmd rest
  in
    find_cmd commands

let rec eval_answer session voc =    
  print_string "> ";
  let answer = Utils.strip (get_line ()) in
    if exec_cmd session voc true answer then (
      print_newline ();
      eval_answer session voc
    ) else ()

let rec ask_yes_no ?(strict=true) q = 
  print_string q;
  print_string "(y, n) ";
  let answer = Utils.strip (get_line ()) in
    if answer = "yes" || answer = "y" then true 
    else if answer = "no" || answer = "n" || not strict then false
    else (print_endline "Please answer 'y' or 'n'."; ask_yes_no q)

let rec ask_yes_no_cmd session voc before_answer q =       
  print_string q;
  print_string "(y, n, some command or :? for help) ";
  let answer = Utils.strip (get_line ()) in
    if answer = "yes" || answer = "y" then true 
    else if answer = "no" || answer = "n" then false
    else (
      if not (exec_cmd session voc before_answer answer) then (
        print_endline "Please answer 'y' or 'n' or give a valid command." 
      );
      print_newline ();
      ask_yes_no_cmd session voc before_answer q
    )

let print_examples l = 
  let rec print l n =
    match l with
        [] -> ()
      | ex :: rest ->
          print_string "  (";
          print_int n;
          print_string ") ";
          print_para ~first_line_indent:false 6 ex;
          print rest (n + 1)
  in print l 1


let rec loop session = 
  let _ = ITRACE "Searching for next vocable to ask" in
  let voc = 
    match VocableSelector.select session with
        Some v -> v
      | None ->
          print_newline ();
          print_endline 
            "Congratulations! No vocable can be selected. This means that";
          print_endline
            "you have to wait a little bit before continuing your training.";
          print_endline
            "Program will exit now.";
          user_exit ()
  in
  let _ = ITRACE "Next vocable found" in
    print_endline ("*** Question ***      " ^
                   "(type your answer, some command or :? for help)");
    print_endline ("                      " ^
                   "          Index: " ^ 
                   ToString.from_option string_of_int ~none:"--" 
                     voc.Vocable.voc_index);
    print_endline ("                      " ^
                   "Asked how often: " ^
                   string_of_int voc.Vocable.voc_timesAsked);
    print_endline ("                      " ^
                   "   Lastly asked: " ^
                   ToString.from_option Utils.string_of_tstamp ~none:"unknown" 
                     voc.Vocable.voc_time);
    print_newline ();
    print_para 2 voc.voc_term;
    print_newline ();
    (match voc.voc_termPronunciation with
         None -> ()
       | Some f -> 
           print_endline 
             "Pronunciation available, type :p for playing it.\n"
    );
    eval_answer session voc;
    print_newline ();
    print_endline "*** Answer ***\n";
    print_para 2 voc.voc_meaning;
    print_newline ();
    (match voc.voc_examples with
         [] -> ()
       | l ->
           print_newline ();
           print_endline "Examples:";
           print_examples l;
           print_newline ();
    );
    (match voc.voc_meaningPronunciation with
         None -> ()
       | Some f -> 
           print_endline 
             "Pronunciation available, type :p for playing it.\n"
    );
    let correct = ask_yes_no_cmd session voc false 
                    "Should I accept your answer? " in
      Update.update_vocable_after_asking session voc correct;
      Statistics.register_answer correct;
      print_newline ();
      if not correct then (
        print_endline "You have now time to learn this vocable.";
        print_endline "Press the ENTER key to continue.";
        ignore (get_line ())
      );
      loop session

let start session =
  start_time := Some (Time.time ());
  loop session
