(*pp ./tracepp *)

open Session


let establish_signal_handlers () = 
  let handle_sigint s = Interactive.user_exit () in
    Sys.set_signal Sys.sigint (Sys.Signal_handle handle_sigint)

let backup_db session = 
  let db_file = session.ses_db_file in
  let make_bak_name v = db_file ^ ".bak" ^ string_of_int v in
  let backup v =
    let fname = make_bak_name v in
    let bakname = make_bak_name (v + 1) in
      if Sys.file_exists fname then (
        ITRACE "Making backup %s -> %s" fname bakname;
        Sys.rename fname bakname
      ) else ()
  in
    List.iter backup [4;3;2;1];
    if Sys.file_exists db_file then
      let bakname = make_bak_name 1 in
      let chan_bak = open_out_bin bakname in
      let chan_db = open_in_bin db_file in
      let rec copy () = 
        output_char chan_bak (input_char chan_db); 
        copy ()
      in 
        (try 
           ITRACE "Making backup %s -> %s" db_file bakname;
           copy ()
         with End_of_file -> ()
        );
        close_in chan_db;
        close_out chan_bak


let main session_file no_interactive =
  let _ = ITRACE "Entering main routine" in
  let session = deserialize_session session_file in
    backup_db session;
    Update.update_vocables session;
    Statistics.display_global_stats session;
    if not no_interactive then (
      print_newline ();
      ITRACE "Entering interactive loop";
      Interactive.start session
    ) else (
      ITRACE "Not entering interactive loop as requested"
    )

let do_global_stats session_file = 
  let _ = ITRACE "Displaying global statistics" in
  let session = deserialize_session session_file in
    Statistics.display_global_stats session


let do_how_many session_file = 
  let _ = ITRACE "Determining number of database items" in
  let session = deserialize_session session_file in
  let db_file = session.ses_db_file in
    Database.with_read db_file
      (fun db ->
         let n = Database.fold_left (fun n _ _ -> n + 1) 0 db in
           print_int n;
           print_newline ()
      )

let action_on_voc session_file s f = 
  let session = deserialize_session session_file in
  let db_file = session.ses_db_file in
    Database.with_read db_file
      (fun db ->
         try 
           let voc = Vocable.deserialize (Database.find db s) in
             f voc
         with
             Not_found -> FATAL "Vocable '%s' not found" s
      )

let do_query session_file q = 
  let _ = ITRACE "Query for %s" q in
    action_on_voc session_file q 
      (fun voc -> print_endline (Vocable.get_meaning voc))

let do_stats session_file s = 
  let _ = ITRACE "Statistics for %s" s in
    action_on_voc session_file s Statistics.display_voc_stats




(************** Argument parsing *******************)

let session_file = ref None
let query_string = ref ""
let how_many = ref false
let stats_string = ref ""
let only_update = ref false
let global_stats = ref false

let set_session_file s = 
  match !session_file with
      None -> session_file := Some s
    | _    -> raise (Arg.Bad ("Only one session file can be given."))

let enable_trace_level s = 
  try
    let l = Trace.level_of_string s in
      Trace.enable_level l
  with
      Invalid_argument _ -> raise (Arg.Bad ("Invalid trace level: " ^ s))

let arg_spec_list = 
  [("-t", Arg.String enable_trace_level, 
    "Enables trace level l. Valid levels: lex, info, debug");
   ("-q", Arg.Set_string query_string,
    "Queries the database for the meaning of the given word and exits.");
   ("-n", Arg.Set how_many,
    "Prints the number of vocables contained in the database and exits.");
   ("-s", Arg.Set_string stats_string,
    "Prints statistics for the given word and exits.");
   ("-u", Arg.Set only_update,
    "Updates the vocable database but does not enter the interactive loop.");
   ("-g", Arg.Set global_stats,
    "Displays a global statistics.")]
   

let _ =
  establish_signal_handlers ();
  Arg.parse arg_spec_list set_session_file
    "Usage: voc [-t level] [-s voc] [-n] [-q voc] [-u] session-file";
  match !session_file with
      None -> FATAL "No session file given."
    | Some s -> 
        (if !query_string <> "" then
           do_query s !query_string
         else if !how_many then
           do_how_many s
         else if !stats_string <> "" then
           do_stats s !stats_string
         else if !global_stats then
           do_global_stats s
         else
           main s !only_update
        );
        exit 0
