(*pp ./tracepp *)

open Lexing
open Printf

type parse_error_data = 
    { ped_filename: string;
      ped_lnum: int;  (* line number *)
      ped_bol: int;   (* characters from beginning of line *)
      ped_next_token: string }
exception Parse_error of parse_error_data

let parse_error_to_string ped = 
    sprintf ("File \"%s\", line %d, character %d:
  Parse error, next token:
  \"%s\"") 
      ped.ped_filename ped.ped_lnum ped.ped_bol ped.ped_next_token

let pos_to_string pos = 
  ("{ pos_fname  = " ^ pos.Lexing.pos_fname ^ ";\n" ^
   "  pos_lnum   = " ^ string_of_int pos.Lexing.pos_lnum ^ ";\n" ^
   "  pos_bol    = " ^ string_of_int pos.Lexing.pos_bol ^ ";\n" ^
   "  pos_cnum   = " ^ string_of_int pos.Lexing.pos_cnum ^ " }")

let parse_error lexbuf =
  let start_pos = Lexing.lexeme_start_p lexbuf
  and end_pos = Lexing.lexeme_end_p lexbuf in
  let fname = start_pos.pos_fname
  and first = start_pos.pos_cnum
  and last = end_pos.pos_cnum - 1 in
  let chan = open_in fname in
  let (lnum, bol) = 
    let rec compute_pos lnum = 
      try
        let start_this_line = pos_in chan in
        let _ = input_line chan in
        let start_next_line = pos_in chan in
          if start_next_line <= first then
            compute_pos (lnum + 1)
          else
            (lnum, first - start_this_line)
      with End_of_file -> (-1, -1)
    in compute_pos 1
  and token = 
    try
      let n = last - first + 1 in
      let buf = String.create n in
        (seek_in chan first;
         really_input chan buf 0 n;
         buf)
    with End_of_file -> ""
  in
    close_in chan;
    { ped_filename = fname;
      ped_lnum = lnum;
      ped_bol = bol;
      ped_next_token = token }

let lexbuf_from_file fname = 
  let chan = open_in fname in
  let lexbuf = Lexing.from_channel chan in
  let new_start_p = { lexbuf.lex_start_p with 
                        Lexing.pos_fname = fname;
                        Lexing.pos_lnum = 1}
  and new_curr_p = { lexbuf.lex_curr_p with 
                       Lexing.pos_fname = fname;
                       Lexing.pos_lnum = 1}
  in (lexbuf.lex_start_p <- new_start_p;
      lexbuf.lex_curr_p <- new_curr_p;
      lexbuf)

let parse fname = 
  let _ = DTRACE "Starting parsing %s" fname in
  let lexbuf = lexbuf_from_file fname in
  let lexer = Lexer.make_lexer () in
  try 
    let res = Parser.main lexer lexbuf in
      DTRACE "Finished parsing %s" fname; res
  with 
      Parsing.Parse_error ->
        raise (Parse_error (parse_error lexbuf))

let debug_lexer () = 
  let lexbuf = lexbuf_from_file "test.voc" in
  let lexer = Lexer.make_lexer () in
  let rec f () = 
    let tok = lexer lexbuf in
      print_string (Token.to_string tok);
      print_newline ();
      match tok with
          Parser.EOF -> ()
        | _   -> f ()
  in f ()

let debug_parser () = 
  let lexbuf = lexbuf_from_file "test.voc" in
  let lexer = Lexer.make_lexer () in
    try
      let vocs = Parser.main lexer lexbuf in
        List.iter (fun v -> 
                     print_string (Vocable.voc_to_string v);
                     print_newline ()) vocs
    with 
        Parsing.Parse_error ->
          print_string (parse_error_to_string (parse_error lexbuf));
          exit 1
      
(*
let _ = debug_parser ()
*)
