(*pp ./tracepp *)

open Session

(* Takes a list of vocables and checks if the list contains
   vocables with the same term.
   Returns 'None' if all terms are unique, 'Some v1 v2' otherwise
   where v1, v2 are the first vocables with the same term that
   were found in l.
*)
let ensure_no_duplicates l = 
  let tbl = Utils.create_hashtbl (List.length l) in
  let rec f = fun l ->
    match l with 
        []      -> None
      | v :: vs ->
          let term = Vocable.get_term v in
            try
              let v2 = Hashtbl.find tbl term in
                Some (v, v2)
            with Not_found -> 
              Hashtbl.add tbl term v;
              f vs
  in f l              

(* Checks if a vocable found in the database is up-to-date compared
   with a vocable with the same term read from a file *)
let is_up_to_date db_voc file_voc =
  let db_voc2 = Vocable.update_index_and_time db_voc None None in
  let file_voc2 = Vocable.update_index_and_time file_voc None None in
    db_voc2 = file_voc2
                                      
(* Merges a vocable from the database with a vocable from a file *)
let merge db_voc file_voc = 
  let i = db_voc.Vocable.voc_index in
  let t = db_voc.Vocable.voc_time in
    Vocable.update_index_and_time file_voc i t

let update_vocables session =
  let _ = ITRACE "updating database from vocable files, session:\n%s"
            (session_to_string session) in
  let find_new_or_modified_vocs db vocables = 
    (* checks for non-existing or modified vocables in the database 
       Returns a tuple (non-existing, modified) *)
    List.fold_left 
      (fun (non_ex, modif) voc ->
         try
           let db_voc = Vocable.deserialize 
                          (Database.find db (Vocable.get_term voc)) in
             if is_up_to_date db_voc voc
             then (non_ex, modif)                         (* unmodified *)
             else (non_ex, merge db_voc voc :: modif)   (* modified *)
         with Not_found ->
           (voc :: non_ex, modif))                        (* non-existing *)
      ([], []) vocables
  in
  let find_old_vocs db vocables = 
    (* checks for vocables that have been deleted from the voc files
       but not from the database *)
    let tbl = Utils.create_hashtbl (List.length vocables) in
    let _ = List.iter (fun voc -> Hashtbl.add tbl (Vocable.get_term voc) ()) 
              vocables in
      Database.fold_left
        (fun l key value ->
           if Hashtbl.mem tbl key
           then l
           else key :: l)
        [] db
  in
  let do_updates db new_vocs mod_vocs old_vocs =
    let add voc = 
      let key = (Vocable.get_term voc) in
        ITRACE "Adding '%s' to database" key;
        Database.add db key (Vocable.serialize voc) 
    in
    let del voc = 
      ITRACE "Deleting '%s' from database" voc;
      Database.remove db voc 
    in
    let modif voc = 
      let key = Vocable.get_term voc in
        ITRACE "Updating '%s' in database" key;
        Database.replace db key
          (Vocable.serialize voc) 
    in
      (List.iter add new_vocs; 
       List.iter del old_vocs; 
       List.iter modif mod_vocs)
  in
  let db_file = session.ses_db_file in
  let voc_files = session.ses_vocable_files in
  let new_vocs, mod_vocs, old_vocs =
    if not (Sys.file_exists db_file) then
      Database.with_write db_file (fun db -> ())
    ;
    Database.with_read db_file
      (fun db ->
         let all_vocables = 
           try 
             let l = List.concat (List.map ParserDriver.parse voc_files) in
               match ensure_no_duplicates l with
                   None          -> l
                 | Some (v1, v2) ->
                     let term = Vocable.get_term v1 in
                     let f1 = Session.get_voc_filename session v1 in
                     let l1 = v1.Vocable.voc_lineno in
                     let f2 = Session.get_voc_filename session v2 in
                     let l2 = v2.Vocable.voc_lineno in
                       FATAL
"Two occurences of vocable '%s' were found in your vocable files.
First in file %s, line %d
Second in file %s, line %d" term f1 l1 f2 l2; []
           with 
               ParserDriver.Parse_error data ->
                 FATAL "%s" (ParserDriver.parse_error_to_string data); []
         in
         let new_vocs, mod_vocs = find_new_or_modified_vocs db all_vocables in
         let old_vocs = find_old_vocs db all_vocables in
           new_vocs, mod_vocs, old_vocs
      )
  in
    Database.with_write db_file
      (fun db ->
         do_updates db new_vocs mod_vocs old_vocs;
      )



let update_vocable_after_asking session voc correct = 
  Database.with_write session.ses_db_file
    (fun db ->
       let new_voc = { voc with 
                         Vocable.voc_timesAsked = voc.Vocable.voc_timesAsked + 1;
                         Vocable.voc_time = Some (Time.time ());
                         Vocable.voc_index = 
                         Some (if correct then 
                                 match voc.Vocable.voc_index with
                                     None -> 2
                                   | Some i -> i + 1
                               else 1) } in
       let _ = 
         let d1, d2 = (match Vocable.diff voc new_voc with
                           Some tuple -> tuple
                         | None       -> ("NO DIFFERENCE", 
                                          "NO DIFFERENCE")) in
           ITRACE "Replacing vocable. Differences:
old: %s
new: %s" d1 d2 
       in
       let key = Vocable.get_term new_voc in
         Database.replace db key (Vocable.serialize new_voc);
    )
