(*pp ./tracepp *)

(*
  Weighted choice algorithm.

  Given:  A set S and for every x in S a real number w (0 <= w). w is
          called the weight of x.
  Result: One element x of S choosen randomly with probability w(x)/T,
          where T is the total of all weights.

  Requirements:
  - |S| <= 50000
  - every x with w(x) < 1/20 is ignored
*)

let mAX_LEN = 50000
let mIN_WEIGHT = 1.0 /. 20.0
let mAX_WEIGHT = 400.0
let nORM_FACTOR = 20.0

let random_initialized = ref false

let init_random () = 
  if not !random_initialized then (
    Random.self_init ();
    random_initialized := true
  ) else ()

(* l is a list of tuples (x, w(x)), f is a function the returns
   for a given x in l a string representation of x (used for debugging,
   error messages). *)
let choose l to_string = 
  (* Checks the all elements in l fulfill the requirements given above,
     normalizes the weights so that w(x) >= 1 holds for every x in S,
     computes the sum of the normalized weights and the length of l.
     Note: The requirements above guarantee that the sum of all
           normalized weights is not greater than  
           50,000 * 400 * 20 = 400,000,000
  *)
  let _ = init_random () in
  let rec check_norm_and_sum l (sum, normed, len) = 
    match l with
        [] -> (sum, normed, len)
      | (x, w) :: rest ->
          if len >= mAX_LEN then (
            FATAL "List for weighted choice is too long. Max: %d" mAX_LEN;
            exit 1
          ) else if w < mIN_WEIGHT then (
            if not (Utils.float_eq w 0.0) then
              WARN ("Weight for '%s' too small. Min: %f, weight: %f. " ^^
                    "Ignoring this vocable")
                    (to_string x) mIN_WEIGHT w
            ;
            check_norm_and_sum rest (sum, normed, len)
          ) else if w > mAX_WEIGHT then (
            FATAL "Weight for '%s' too big. Max: %f, weight: %f"
                     (to_string x) mAX_WEIGHT w;
            exit 1
          ) else
            let nw = int_of_float (ceil (w *. nORM_FACTOR)) in
              check_norm_and_sum rest (sum + nw, (x, nw) :: normed, len + 1)
  in
  (* Selects one item of l randomly. 'sum' is the sum of all weights in l.
     The sum must be less that the greatest integer value.
     Every weight in l must be an integer and all weights must be 
     greater than zero. *)
  let choose_randomly l sum =
    let _ = DTRACE "sum: %d" sum in
    let rand = (Random.int sum) + 1 in
    let _ = DTRACE "random number: %d" rand in
    let rec f l t =  (* l is the remaining list, 
                        t is the sum of weights so far *)
      match l with
          [] -> FATAL "End of list reached"; exit 1
        | (x, w) :: rest ->
            let newt = t + w in
              if rand <= newt then
                x
              else
                f rest newt
    in f l 0            
  in
  let (sum, normed, _) = check_norm_and_sum l (0, [], 0) in
    if sum = 0 then (
      (* This is the rare case where no vocable can be selected *)
      DTRACE "sum is 0, no vocable can be selected";
      None
    ) else 
      Some (choose_randomly normed sum)

